###########################################################
## Poisson Regression Changepoint Model
## using Gibbs sampling algorithm
###########################################################

## Note that this R code is not optimized for fast computation.
## data augmentation method is done using the algorithm proposed by
## Sylvia Fruhwirth-Schnatter and Helga Wagner (2006) "Auxiliary Mixture Sampling for Parameter-driven Models of Time Series of Small Counts with Applications to State Space Modelling", Biometrika

## 01/31/2007
## Written by Jong Hee Park

## Model
## y ~ Poisson(mu_i)
## mu_i = X%*%beta_k
## beta_k = {beta_1, ... , beta_(m+1)}
## parameter = {beta_k, P}

## prior
## beta ~ N(b0, B0)
## Pii  ~ Beta()

#########################################################
"MCMCpoissonRegChangepoint"<-
#########################################################
  function(formula, data=parent.frame(), m=1, b0, B0,
           burnin = 1000, mcmc = 1000, thin = 1, verbose = 0, a = NULL, b = NULL,
           seed = NA, marginal.likelihood = c("none", "Chib95"), ...) {

    ## Arguments

    ## m : the number of changepoint
    ## b0 and B0: prior for beta. NO DEFAULT values
    ## a and b: beta prior for transition probabilities
    ## By default, the expected duration is computed and
    ## corresponding a and b values are assigned. The expected
    ## duration is the sample period divided by the number of states.
    ## ex) 300 years and 2 changepoints (3 states)
    ##        expected.duration <- 300/(2+1)
    ##        b <- 0.1
    ##        a <- b*expected.duration

    ## Mixture of normals
    wr  <-  c(0.2294, 0.2590, 0.2480, 0.1525, 0.0472)
    mr  <-  c(0.0982, -1.5320, -0.7433, 0.8303, -3.1428)
    sr  <-  sqrt(c(0.2401, 1.1872, 0.3782, 0.1920, 3.2375))

    ## form response and model matrices
    holder <- MCMCpack:::parse.formula(formula, data)
    y <- holder[[1]]
    X <- holder[[2]]
    xnames <- holder[[3]]
    K <- ncol(X)    # number of covariates
    n <- length(y)
    n.arrival<-  y + 1
    NT      <-  max(n.arrival)   # maximum value of nt
    tot.comp<-  n + sum(y)       # total sum of components
    ns <- m + 1 # number of states

    # check iteration parameters
    MCMCpack:::check.mcmc.parameters(burnin, mcmc, thin)
    totiter <- mcmc + burnin
    cl <- match.call()

    ## prior
    A0 <- trans.mat.prior(m=m, n=n, a=a, b=b)
    if(is.null(b0)||is.null(B0))
    stop("You need to specify the prior for beta!")
    B0inv <- solve(B0)

    ## get marginal likelihood argument
    marginal.likelihood  <- match.arg(marginal.likelihood)

    ## starting values for beta and var
    beta <-  rmvnorm(ns, b0, B0)
    P  <-  trans.mat.prior(m=m, n=n, a=0.9, b=0.1)
    tau <- tau.initial(y, tot.comp)
    component  <-  round(runif(tot.comp, 1, 5))
    Yt      <-  as.list(rep(NA, n))
    Sigma   <-  as.list(rep(NA, n))
    y.tilde <-  rep(NA, n)

    ## storage matrices
    beta.store  <-  array(NA, c(ns, mcmc/thin, K))
    component.store <-  matrix(NA, mcmc/thin, tot.comp)
    tau.store   <-  matrix(NA, mcmc/thin, tot.comp)

    P.store     <-  matrix(NA, mcmc/thin, ns^2)
    ps.store    <-  matrix(0, n, ns)
    s.store     <-  matrix(NA, mcmc/thin, n)
    py          <-  rep(0, ns)

####################################################################
    ## MCMC iteration starts!
####################################################################

    for (iter in 1:totiter){

      count   <-  0
      for (t in 1:n){
        tn      <-  y[t]
        count   <-  count + tn + 1
        itnt    <-  (count-tn):count
        Yt[[t]] <-  log(tau[itnt])- mr[component[itnt]]

        if (length(itnt)==1) Sigma[[t]]  <-  matrix(sr[component[itnt]]^2,1,1)
        else   {Sigma[[t]]  <-  (sr[component[itnt]])^2}
        y.tilde[t]  <-  sum(Yt[[t]]/Sigma[[t]])
      }

      ## Step 1: Sample S
      state.out <- Poisson.reg.state.sampler(m=m, y=y, X=X, beta=beta, P=P)
      s <- state.out$s
      ps<- state.out$ps

      ## Step 2: Sample Beta
      count  <-  0
      for (j in 1:ns){
        ej      <-  as.numeric(s==j)
        yj      <-  matrix(y.tilde[ej==1],,1)
        Xj      <-  matrix(X[ej==1,],,K)
        nt      <-  sum(ej)
        count   <-  count + nt
        itnt    <-  (count-nt+1):count
        wi      <-  sapply(itnt, function(t){sum(1/Sigma[[t]])})
        Xjw     <-  wi*Xj
        Bn  <-  chol2inv(chol(B0inv + (t(Xj)%*%Xjw)))
        bn  <-  Bn%*%(B0inv%*%b0 - (t(Xj)%*%yj))
        beta[j,]  <-  t(rmvnorm(1, bn, Bn))
      }

      ## Step 3: Sample P (independent of beta and y)
      switch  <-  switchg(s)
      for (j in 1:(m+1)){
        switch1 <-  A0[j,] + switch[j,]
        pj      <-  rdirichlet.cp(1, switch1)
        P[j,]   <-  pj
      }

      ## Step 4: Sample tau
      alpha <- rep(0, n)
      tau.out <- tau.comp.sampler(y, X, beta, alpha, s, tot.comp, wr, mr, sr)
      tau <- tau.out$tau
      component <- tau.out$component

      ## store simulated parameters
      if (iter > burnin && iter %%thin == 0){
        beta.store[,(iter-burnin)/thin,]        <-  beta
        component.store[(iter-burnin)/thin,]    <-  component
        tau.store[(iter-burnin)/thin,]          <-  tau
        P.store[(iter-burnin)/thin,]            <-  as.vector(t(P))
        s.store[(iter-burnin)/thin,]            <-  s
        ps.store                                <-  ps.store + ps
      }

      ## report
    if(verbose > 0 && iter%%verbose == 0){
      cat("----------------------------------------------",'\n')
      cat("iteration = ", iter, '\n')
      for(i in 1:(m+1))
        cat(paste("", beta[i,]), fill=TRUE, labels=paste("beta(",i,"):", sep=""), sep=",")
      cat("Transition Matrix", '\n')
      for(i in 1:(m+1))
        cat(paste("", P[i,]), fill=TRUE, labels=paste("{",i,"}:", sep=""), sep=",")
    }

    }
    ## end of MCMC run

    ## marginal likelihood calculation if Chib95 == TRUE
    if (marginal.likelihood == "Chib95"){

      ## Bayes Factor Calculation Starts
      beta.st     <-  sapply(c(1:K), function(i){apply(beta.store[,,i], 1, mean)})
      P.vec.st    <-  apply(P.store, 2, mean)
      P.st        <-  t(matrix(P.vec.st, m+1, m+1))
      pdf.beta.store <- matrix(NA, mcmc/thin, ns)
      pdf.P.store <-  matrix(NA, mcmc/thin)

      ## 1. pdf.beta
      for (g in 1:(mcmc/thin)){
        component.g <-  component.store[g,]
        tau.g <-  tau.store[g,]
        count   <-  0
        for (t in 1:n){
          tn      <-  y[t]
          count   <-  count + tn + 1
          itnt    <-  (count-tn):count
          Yt[[t]] <-  log(tau.g[itnt])- mr[component.g[itnt]]
          if (length(itnt)==1) Sigma[[t]]  <-  matrix(sr[component.g[itnt]]^2,1,1)
          else   {Sigma[[t]]  <-  (sr[component.g[itnt]])^2}
          y.tilde[t]  <-  sum(Yt[[t]]/Sigma[[t]])
        }
        count   <-  0
        for (j in 1:ns){
          ej      <-  as.numeric(s.store[g,]==j)
          yj      <-  matrix(y.tilde[ej==1],,1)
          Xj      <-  matrix(X[ej==1,],,K)
          nt      <-  sum(ej)
          count   <-  count + nt
          itnt    <-  (count-nt+1):count
          wi      <-  sapply(itnt, function(t){sum(1/Sigma[[t]])})
          Xjw     <-  wi*Xj
          Bn      <-  chol2inv(chol(B0inv + (t(Xj)%*%Xjw)))
          bn      <-  Bn%*%(B0inv%*%b0 - (t(Xj)%*%yj))
          pdf.beta.store[g, j]    <-  dmvnorm(beta.st[j,], bn, Bn)
        }
      }
      log.pdf.beta <- log(prod(apply(pdf.beta.store, 2, mean)))

      ## 2. pdf.P: p(P.st|y, beta.st, S)
      pdf.P.temp.store  <-  rep(NA, m+1)
      for (g in 1:(mcmc/thin)){
        state.out <- Poisson.reg.state.sampler(m=m, y=y, X=X, beta=beta.st, P=P)
        s <- state.out$s
        switch  <-  switchg(s)
        for (j in 1:(m+1)){
          switch1 <-  A0[j,] + switch[j,]
          P[j,] <-  rdirichlet.cp(1, switch1)
          pdf.P.temp.store[j] <- ddirichlet.cp(P.st[j,], switch1)
        }
        pdf.P.store[g]<- prod(pdf.P.temp.store)
      }

      log.pdf.P     <- log(mean(pdf.P.store))

      ## likelihood
      F   <-  matrix(NA, n, ns)
      like<-  rep(NA, n)
      pr1 <-  c(1,rep(0, m))
      for (t in 1:n){
        py  <-  sapply(c(1:ns), function(i){poisson.reg.pdf(y[t], X[t,], beta.st[i,])})
        if(t==1) {pstyt1 = pr1}
        else {pstyt1 <- F[t-1,]%*%P.st}
        unnorm.pstyt <- pstyt1*py
        pstyt   <- unnorm.pstyt/sum(unnorm.pstyt)
        F[t,]   <-  pstyt
        like[t] <-  sum(unnorm.pstyt)
      }
      loglik  <-  sum(log(like))

      ## prior ordinates
      prior.beta     <-  sum(sapply(c(1:ns), function(i){log(dmvnorm(beta.st[i,], b0, B0))}))
      prior.P        <-  sum(sapply(c(1:ns), function(i){log(ddirichlet.cp(P.st[i,], A0[i,]))}))

      ## Marginal Likelihood                    ##
      numerator   <-  loglik + prior.beta + prior.P
      denominator <-  log.pdf.beta + log.pdf.P
      logmarglike <-  numerator - denominator

      ## print
      if(verbose > 0){
        cat("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n")
        cat("Log Marginal Likelihood\n")
        cat("-------------------------------------------------",'\n')
        cat("log(marglike)= ",      logmarglike, '\n')
        cat("log(likelihood)= ",    loglik, '\n')
        cat("------ Display marginal likelihood components for error check -------",'\n')
        cat("log(prior.beta)= ",    prior.beta, '\n')
        cat("log(prior.P)= ",       prior.P, '\n')
        cat("log(pdf.beta)= ",      log.pdf.beta, '\n')
        cat("log(pdf.P)= ",         log.pdf.P, '\n')
        cat("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n")
      }
    } ## end of marginal likelihood calculation

    else {logmarglike <- loglik <- NULL}

    ##  return output
    beta.mcmc <- matrix(NA, mcmc/thin, K*ns)
    for (i in 1:ns) beta.mcmc[,((i-1)*K + 1):((i-1)*K + K)] <- beta.store[i,,]
    output <- as.mcmc(beta.mcmc)
    varnames(output)<- paste(xnames, "_regime_", rep(1:ns, each=K), sep = "")
    attr(output, "title") <- "MCMCpoissonRegChangepoint Samples"
    attr(output, "y") <- y
    attr(output, "X") <- X
    attr(output, "m") <- m
    attr(output, "call") <- cl
    attr(output, "logmarglike") <- logmarglike
    attr(output, "loglik") <- loglik
    attr(output, "prob.state") <- ps.store/(mcmc/thin)
    attr(output, "s.store") <- s.store

    return(output)

  }## end of MCMC function

## ## load two libraries before run the code
## require(MCMCpack)
## require(mvtnorm)

## ## example
## set.seed(11119)
## n <- 200
## x1 <- runif(n, 0, 0.5)
## true.beta1 <- c(0.5,  1)
## true.beta2 <- c(-1,  3)
## true.beta3 <- c(1,  2)
## true.s <- rep(1:2, each=n/2) ## true state changes

## mu1 <- exp(cbind(1, x1[true.s==1])%*%true.beta1)
## mu2 <- exp(cbind(1, x1[true.s==2])%*%true.beta2)
## y <- as.ts(c(rpois(n/2, mu1), rpois(n/2, mu2)))
## formula = y ~ x1

## ## run the example
## start <- Sys.time()
## model1 <-  MCMCpoissonRegChangepoint(formula, m=1, b0= rep(0, 2), B0 = 2*diag(2), burnin = 10, mcmc = 10, verbose = 10, marginal.likelihood = "Chib95")
## model2 <-  MCMCpoissonRegChangepoint(formula, m=2, b0= rep(0, 2), B0 = 2*diag(2), burnin = 10, mcmc = 10, verbose = 10, marginal.likelihood = "Chib95")
## model3 <-  MCMCpoissonRegChangepoint(formula, m=3, b0= rep(0, 2), B0 = 2*diag(2), burnin = 10, mcmc = 10, verbose = 10, marginal.likelihood = "Chib95")
## model4 <-  MCMCpoissonRegChangepoint(formula, m=4, b0= rep(0, 2), B0 = 2*diag(2), burnin = 10, mcmc = 10, verbose = 10, marginal.likelihood = "Chib95")
## model5 <-  MCMCpoissonRegChangepoint(formula, m=5, b0= rep(0, 2), B0 = 2*diag(2), burnin = 10, mcmc = 10, verbose = 10, marginal.likelihood = "Chib95")
## print(Sys.time()-start)

## BayesFactor(model1, model2, model3, model4, model5)



