# PoissonRegPosterior-V4.R
#
# Posterior analyis for the changepoint models
#
# Patrick T. Brandt
#
# 20090406 : Initial version
# 20090420 : Revised to work with the split version of the data and
#            code in the PoissonCP*.R files
# 20090513 : Analyzes the posterior for the subset of covariates for
#            logistical outcomes and people v. property only.
# 20090922 : Updated to make a table of the first differences effects.

library(zoo)
library(MCMCpack)
library(mvtnorm)
source("utility.R")
source("MCMCpoissonStatespace.R")
source("MCMCpoissonRegChangepoint.R")

# Load the posterior sample data
load("PoissonCP1-V4.RData")
load("PoissonCP2-V4.RData")

# Make tables of the log Bayes factors
library(xtable)

system("rm ../logBF-V4.tex")
tmp <- Officials.BF$BF.log.mat
rownames(tmp) <- colnames(tmp) <- 1:4
print(xtable(tmp,
       caption="Log Bayes factors for number of changepoints in the Officials series",
       label="tab:OfficialsBF"), file="../logBF-V4.tex", append=TRUE)

tmp <- Military.BF$BF.log.mat
rownames(tmp) <- colnames(tmp) <- 1:4
print(xtable(tmp,
       caption="Log Bayes factors for number of changepoints in the Military series",
       label="tab:MilitaryBF"), file="../logBF-V4.tex", append=TRUE)

tmp <- Business.BF$BF.log.mat
rownames(tmp) <- colnames(tmp) <- 1:4
print(xtable(tmp,
       caption="Log Bayes factors for number of changepoints in the Business series",
       label="tab:BusinessBF"), file="../logBF-V4.tex", append=TRUE)

tmp <- PrivateParties.BF$BF.log.mat
rownames(tmp) <- colnames(tmp) <- 1:4
print(xtable(tmp,
       caption="Log Bayes factors for number of changepoints in the Private Parties series",
       label="tab:PrivatePartiesBF"), file="../logBF-V4.tex", append=TRUE)


# Now summarize some of the posterior results for the best fitting
# models

Officials.post <- out.reg1[[2]]
Military.post <- out.reg1[[7]]
Business.post <- out.reg2[[2]]
PrivateParties.post <- out.reg2[[7]]

# First, plot where the changepoints are.  Note that we are going to
# do this like a technical analysis plot so we can see the uncertainty
# as well.  Do this with a custom function so we can control more than
# what is done in plotState() and plotChangepoint() in MCMCpack /
# JHP's code.

plotChange <- function(x, verbose=TRUE, name=NULL)
{
    # Setup data for plot
    m <- attr(x, "m")
    out <- zoo(cbind(attr(x, "y"), attr(x, "prob.state")),
               as.yearmon(dimnames(attr(x, "y"))[[1]], "%b %Y"))

    # Compute the location of the changepoints
    cp.means <- rep(NA, m+1)      # regime means
    pr.st <- diff(out[,3:(m+2)])  # regime probs
    pr.st[pr.st<0] <- 0           # zeros
    pr.cum <- cumsum(pr.st)       # cum probs
    cp.store <- rep(NA, m)        # storage for CPs dates
    cp.idx <- rep(NA, m)

    for(i in 1:m)
    {
        tmp <- pr.cum[,i]
        cp.idx[i] <- which(tmp>0.5)[1]-1
        cp.store[i] <- as.numeric(index(tmp[tmp>0.5])[1])
    }

    # Now begin plotting the first two panels: data and the regimes
    par(mfrow=c(2+m, 1), mai=c(0.4, 0.6, 0.1, 0.05), xpd=TRUE)
    plot(out[,1], xlab="", ylab=name)
    plot(out[,2:(m+2)], plot.type="single", col=1:(m+1),
         xlab="", lwd=2, ylab="Regime Probabilities")

    # Add the text for the means
    cp.start <- c(1, cp.idx+1)
    cp.end <- c(cp.idx, nrow(out))

    for(i in 1:(m+1))
    {
        cp.mn <- mean(out[cp.start[i]:cp.end[i],1])
        tp <- tsp(as.ts(out))
        cp.mid <- tp[1] + cp.start[i]/tp[3] + (0.5*(cp.end[i]-cp.start[i])/tp[3])
        text(x=cp.mid, y=0.5, round(cp.mn, 1), cex=1.5)
    }

    # Now plot the densities for the changepoints
    par(xpd=FALSE)
    for(i in 1:m)
    {
        plot(pr.st[,i], type="h", main="", xlab="",
             ylab=paste("Pr(Regime ", i+1, ")", sep=""))
        abline(v = cp.store[i], lty = 3, lwd=2, col = i+1)
    }

    if (verbose == TRUE) {
        cat("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n")
        cat("Expected changepoint(s): \n")
        for (i in 1:(m)) print(cp.store[i])
        cat("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n")
    }
}


# Same as the above plotting function, but it only plots the first two
# panels.
plotChange2 <- function(x, verbose=TRUE, name=NULL)
{
    # Setup data for plot
    m <- attr(x, "m")
    out <- zoo(cbind(attr(x, "y"), attr(x, "prob.state")),
               as.yearmon(dimnames(attr(x, "y"))[[1]], "%b %Y"))

    # Compute the location of the changepoints
    cp.means <- rep(NA, m+1)      # regime means
    pr.st <- diff(out[,3:(m+2)])  # regime probs
    pr.st[pr.st<0] <- 0           # zeros
    pr.cum <- cumsum(pr.st)       # cum probs
    cp.store <- rep(NA, m)        # storage for CPs dates
    cp.idx <- rep(NA, m)

    for(i in 1:m)
    {
        tmp <- pr.cum[,i]
        cp.idx[i] <- which(tmp>0.5)[1]-1
        cp.store[i] <- as.numeric(index(tmp[tmp>0.5])[1])
    }

    # Now begin plotting the first two panels: data and the regimes
    par(mfrow=c(2, 1), mai=c(0.5, 0.85, 0.1, 0.1), xpd=TRUE)
    plot(out[,1], xlab="", ylab=name)
    plot(out[,2:(m+2)], plot.type="single", col=1:(m+1),
         xlab="", lwd=2, ylab="Regime Probabilities")

    # Add the text for the means
    cp.start <- c(1, cp.idx+1)
    cp.end <- c(cp.idx, nrow(out))

    for(i in 1:(m+1))
    {
        cp.mn <- mean(out[cp.start[i]:cp.end[i],1])
        tp <- tsp(as.ts(out))
        cp.mid <- tp[1] + cp.start[i]/tp[3] + (0.5*(cp.end[i]-cp.start[i])/tp[3])
        text(x=cp.mid, y=0.5, round(cp.mn, 1), cex=1.25)
    }

    if (verbose == TRUE) {
        cat("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n")
        cat("Expected changepoint(s): \n")
        for (i in 1:(m)) print(cp.store[i])
        cat("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n")
    }
}


# Now generate the plots
pdf(file="../OfficialsCP-V4.pdf", width=6, height=8)
plotChange(Officials.post, name="Officials")
dev.off()

pdf(file="../MilitaryCP-V4.pdf", width=6, height=8)
plotChange(Military.post, name="Military")
dev.off()

pdf(file="../BusinessCP-V4.pdf", width=6, height=8)
plotChange(Business.post, name="Business")
dev.off()

pdf(file="../PrivatePartiesCP-V4.pdf", width=6, height=8)
plotChange(PrivateParties.post, name="Private Parties")
dev.off()

# Now do the two panel ones.
pdf(file="../OfficialsCP2-V4.pdf", width=6, height=4)
plotChange2(Officials.post, name="Officials")
dev.off()

pdf(file="../MilitaryCP2-V4.pdf", width=6, height=4)
plotChange2(Military.post, name="Military")
dev.off()

pdf(file="../BusinessCP2-V4.pdf", width=6, height=4)
plotChange2(Business.post, name="Business")
dev.off()

pdf(file="../PrivatePartiesCP2-V4.pdf", width=6, height=4)
plotChange2(PrivateParties.post, name="Private Parties")
dev.off()

## # Now plot the posterior coefficents.  Will use a new function here to
## # we can then utilize some coda fu.

## summary.change.coef <- function(x, name=NULL)
## {
##     # Get some constants
##     n <- attr(x, "mcpar")[2]
##     k <- ncol(attr(x, "X"))
##     m <- attr(x, "m")

##     # Set this up so we can do a density overplot
##     mc <- NA
##     for(i in 1:(m+1))
##     {
##         if(i==1) {
##             mc <- x[,1:k]
##             g <- rep(1, n)
##         } else {
##             mc <- rbind(mc, x[,(((i-1)*k)+1):(((i-1)*k)+k)])
##             g <- c(g, rep(i, n))
##         }
##     }
##     mc <- mcmc(cbind(mc,g))
##     vname <- sub("_regime_1", "", colnames(mc), perl=TRUE)
##     colnames(mc) <- c(vname)

##     print(densityplot(mc[,1:ncol(mc)-1], groups=g,
##                       col=1:(m+1), plot.points="",
##                       lwd=2))

##     tmp <- summary(x)
##     # Matrix of the means and 95% CI
##     mean.coef <- matrix(tmp$statistics[,1],k)
##     ci.coef <- tmp$quantiles[,c(1,5)]
##     ci.coef <- array(t(ci.coef), c(2, k, m+1))

##     # Now may a pretty latex table
##     cat(paste("\\begin{table}[tb]\n \\begin{tabular}{cccc}",
##               "\n", sep=""))
##     for(i in 1:k)
##     {
##         cat(paste(vname[i], "& ", sep = " "))
##         for(j in 1:(m+1))
##         {
##             if(j<m+1)
##             {
##                 cat(paste(round(mean.coef[i,j],2), "& ", sep=" "))

##             } else {
##                 cat(paste(round(mean.coef[i,j],2), "\\\\", sep=" "))

##             }
##         }
##         cat("\n & ")
##         for(j in 1:(m+1))
##         {
##             if(j<m+1)
##             {
##                 cat(paste("[", round(ci.coef[1,i,j],2), ", ",
##                           round(ci.coef[2,i,j],2), "] & ", sep=""))
##             } else {
##                 cat(paste("[", round(ci.coef[1,i,j],2), ", ",
##                           round(ci.coef[2,i,j],2), "] \\\\", sep=""))
##             }
##         }
##         cat("\n")
##     }

##     cat("\\end{tabular}\n")
##     cat("\\caption{", name, "}\n \\end{table}\n", sep="")
## }

## # Now summarize the results

## pdf(file="../Officialscoef-V4.pdf")
## summary.change.coef(Officials.post, "Officials")
## dev.off()

## pdf(file="../Militarycoef-V4.pdf")
## summary.change.coef(Military.post, "Military")
## dev.off()

## pdf(file="../Businesscoef-V4.pdf")
## summary.change.coef(Business.post, "Business")
## dev.off()

## pdf(file="../PrivatePartiescoef-V4.pdf")
## summary.change.coef(PrivateParties.post, "Private Parties")
## dev.off()


# Now estimate some marginal effects, first differences, etc.

summary.effects <- function(x)
{
    X <- attr(x, "X")
    m <- attr(x, "m")
    k <- ncol(x)/(m+1)

    N <- attr(x, "mcpar")[2]

    # Percentage changes
    per.change <- mcmc(as.matrix((exp(x)-1)*100))
    colnames(per.change) <- paste(rep(colnames(X), m+1), rep(1:(m+1),
                                                             each=k), sep="_")

    # First differences -- regime specific, accounting for the
    # changing data that goes into the computation!

    # Function gets the prediction for one set of draws, accounting for the
    # changepoints and their uncertainty.  Uses a bit of apply() foo.
    # Note also we skip doing the computation for the intercepts.

    tmp <- sapply(1:N, function(i)
              {
                  fd.out <- matrix(0, (m+1), (k-1))
                  for(j in 1:(m+1))
                  {
                 # Mean prediction
                  pred <- (exp(tcrossprod(matrix(X[attr(x,"s.store")[i,]==j,], ncol=k),
                                          matrix(x[i,(((j-1)*k)+1):(((j-1)*k)+k)], nrow=1))))

                 # One unit increases in each variable, one at a time.
                  for (h in 2:k)
                  {
                      X1 <- matrix(X[attr(x, "s.store")[i,]==j,], ncol=k)
                      X1[,h] <- X1[,h]+1

                      pred1 <- (exp(tcrossprod(X1,
                                               matrix(x[i,(((j-1)*k)+1):(((j-1)*k)+k)], nrow=1))))

                      fd.out[j,(h-1)] <- mean(pred1-pred) # mean
                                        # effect, one regime, one draw
                  }
              }
              return(fd.out)
          })
    fd <- mcmc(t(tmp))
    colnames(fd) <- paste(rep(colnames(X)[2:k], m+1), rep(1:(m+1), each=k-1), sep="_")

    # Now compute marginal effects
    tmp <- sapply(1:N, function(i)
              {
                  me.out <- matrix(0, (m+1), (k))
                  for(j in 1:(m+1))
                  {
                 # Mean prediction
                  me.out[j,] <- apply(exp(tcrossprod(matrix(X[attr(x, "s.store")[i,]==j,],ncol=k),
                                                     matrix(x[i,(((j-1)*k)+1):(((j-1)*k)+k)], nrow=1)))%*%
                                                         matrix(x[i,(((j-1)*k)+1):(((j-1)*k)+k)],nrow=1),
                                                         2, mean)
              }

                  return(me.out)
              })
    me <- mcmc(t(tmp))
    colnames(me) <- paste(rep(colnames(X), m+1), rep(1:(m+1),
                                                     each=k), sep="_")

    out <- list(pc=per.change, fd=fd, me=me)
    attr(out, "m") <- m
    return(out)

}

# Sumamrize the first differences with some lattice plots

summary.fd <- function(x, name, ...)
{
    n <- attr(x$fd, "mcpar")[2]
    m <- attr(x, "m")
    k <- ncol(x$pc)/(m+1)

    # Cheat this using an MCMC list density plot
    mc <- vector("list", m+1)
    mc[[1]] <- x$fd[,1:(k-1)]
    vname <- sub("_1", "", colnames(mc[[1]]), perl=TRUE)
    colnames(mc[[1]]) <- vname

    for(i in 2:(m+1))
    {

        mc[[i]] <- x$fd[,(((i-1)*(k-1))+1):(((i-1)*(k-1))+(k-1))]
        colnames(mc[[i]]) <- vname
    }

    class(mc) <- "mcmc.list"
    print(densityplot(mc, col=1:(m+1), plot.points="",
                      lwd=2, ...))

    # Now dump out a table of the same things
    tmp <- summary(x$fd)
    # Matrix of the means and 95% CI
    mean.fd <- matrix(tmp$statistics[,1],k-1)
    ci.fd <- tmp$quantiles[,c(1,5)]
    ci.fd <- array(t(ci.fd), c(2, k-1, m+1))

    # Now may a pretty latex table
    cat(paste("\\begin{table}[tb]\n \\begin{tabular}{cccc}",
              "\n", sep=""))
    for(i in 1:(k-1))
    {
        cat(paste(vname[i], "& ", sep = " "))
        for(j in 1:(m+1))
        {
            if(j<m+1)
            {
                cat(paste(round(mean.fd[i,j],2), "& ", sep=" "))

            } else {
                cat(paste(round(mean.fd[i,j],2), "\\\\", sep=" "))

            }
        }
        cat("\n & ")
        for(j in 1:(m+1))
        {
            if(j<m+1)
            {
                cat(paste("[", round(ci.fd[1,i,j],2), ", ",
                          round(ci.fd[2,i,j],2), "] & ", sep=""))
            } else {
                cat(paste("[", round(ci.fd[1,i,j],2), ", ",
                          round(ci.fd[2,i,j],2), "] \\\\", sep=""))
            }
        }
        cat("\n")
    }

    cat("\\end{tabular}\n")
    cat("\\caption{", name, "}\n \\end{table}\n", sep="")

}


# Now generate the first differences
Officials.effects <- summary.effects(Officials.post)
Military.effects <- summary.effects(Military.post)
Business.effects <- summary.effects(Business.post)
PrivateParties.effects <- summary.effects(PrivateParties.post)

pdf(file="../Officialsfd-V4.pdf", width=6.5, height=2.25)
summary.fd(Officials.effects, "Officials First Differences",
           layout=c(5,1), aspect=1, from=-5, to=11, xlim=c(-5,10),
           scales=list(cex=0.65, tick.number=4))
dev.off()

pdf(file="../Militaryfd-V4.pdf", width=6.5, height=2.25)
summary.fd(Military.effects, "Military First Differences",
           layout=c(5,1), aspect=1, from=-2, to=4, xlim=c(-3,4),
           scales=list(cex=0.65, tick.number=4))
dev.off()

pdf(file="../Businessfd-V4.pdf", width=6.5, height=2.25)
summary.fd(Business.effects, "Business First Differences",
           layout=c(5,1), aspect=1, from=-2, to=5, xlim=c(-2,5),
           scales=list(cex=0.65, tick.number=4))

dev.off()

pdf(file="../PrivatePartiesfd-V4.pdf", width=6.5, height=2.25)
summary.fd(PrivateParties.effects, "Private Parties First Differences",
           layout=c(5,1), aspect=1, from=-5, to=7, xlim=c(-3,3),
           scales=list(cex=0.65, tick.number=4))
dev.off()

# Now make the output of the 68% confidence region
print(Officials.fd68 <- summary(Officials.effects$fd, quantile=c(0.16, 0.5, 0.84)))
print(Military.fd68 <- summary(Military.effects$fd, quantile=c(0.16, 0.5, 0.84)))
print(Business.fd68 <- summary(Business.effects$fd, quantile=c(0.16, 0.5, 0.84)))
print(PrivateParties.fd68 <- summary(PrivateParties.effects$fd, quantile=c(0.16, 0.5, 0.84)))

# Now let's make a table of signed effects for these -- we'll be able
# to assemble this into a table in LaTeX

signs <- function(x)
{
    x1 <- x$quantiles
    k <- nrow(x1)
    names <- rownames(x1)

    for(i in 1:k)
    {
        cat(names[i], " & ")
        if(x1[i,1] > 0 & x1[i,3]>0) cat("+ & \n")
        if(x1[i,1] < 0 & x1[i,3]<0) cat("- & \n")
        if(x1[i,1] < 0 & x1[i,3]>0) cat("0 & \n")
    }
}


signs(Officials.fd68)
signs(Military.fd68)
signs(Business.fd68)
signs(PrivateParties.fd68)

# Quit nicely

q(save="no")

